#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 15 09:59:25 2022

@author: nurmelaj
"""

import time
from datetime import datetime
import xarray as xr
import numpy as np
import pandas as pd
from scipy.optimize import least_squares
import warnings
warnings.filterwarnings('ignore', category=RuntimeWarning)

import ddeq
from ddeq.mcmc_tools import AM_MC
from ddeq.misc import init_results_dataset
from ddeq.divergence_model import model_residual,gridded_flux,compute_divergence_map
from ddeq.wind import read_gnfra_wind

def estimate_emission(data_path, wind_folder,
                      ignore = ['Belchatow','Boxberg','Chvaletice','Hamburg','Pocerady',
                                'Prague','Prunerov','Schwarze Pumpe','Staudinger'],
                      lon_km=50, lat_km=50, grid_reso=5,
                      sats=['0000','0805','1610'], wind_product='ERA5',
                      co2_noise_scenario='low', co2_cloud_threshold=0.01,
                      no2_noise_scenario='low', no2_cloud_threshold=0.30,
                      start_date='2015-01-01', end_date='2015-12-24',
                      AM_options={'warmup':10000,'samples':90000},
                      trace_gases=['CO2']):
    """ 
    Estimates the NOx and CO2 emissions for one source over the chosen time series
    by computing the divergence map using the SMARTCARB data with the chosen satellites
    and fitting a peak fitting model.
    """
    # Process start time
    estimation_start = time.process_time()
    # Dates as datetime objects
    start_date = datetime.strptime(start_date,'%Y-%m-%d') if isinstance(start_date, str) else start_date
    end_date = datetime.strptime(end_date,'%Y-%m-%d') if isinstance(end_date, str) else end_date
    # Source xarray object
    sources = ddeq.misc.read_point_sources()
    sources = sources.where(xr.DataArray([name not in ignore for name in sources['source'].data],dims='source'),drop=True)
    
    dates = pd.date_range(start_date,end_date)
    # Validate supported trace gases
    for gas in trace_gases:
        if gas not in ['CO2','NO2']:
            raise ValueError(f'Trace gas {gas} not supported')
    if wind_product not in ['ERA5','COSMO']:
        raise ValueError(f'Unknown wind product {wind_product}')
    
    results_ds = init_results_dataset(sources, trace_gases, units='Mt/y',method='divergence')
    for source_name in sources['source'].data:
        #emissions = pd.read_csv(f'{ddeq_path}/ddeq/data/ts_{gas}_emissions.dat',delim_whitespace=True,parse_dates=[['date','UTC']])
        #emissions.set_index('date_UTC',inplace=True)
        for gas in trace_gases:
            print(f"Computing divergence map for source '{source_name}'...")
            start_time = time.process_time()
            # First date, ignore timestamp
            date = dates[0].date()
            source_xr = sources.loc[{'source':source_name}]
            center_lon, center_lat = float(source_xr['lon_o'].data), float(source_xr['lat_o'].data)
            # Regular km grid around the source
            longrid, latgrid, longrid_km, latgrid_km = ddeq.misc.grids(center_lon,center_lat,lon_km,lat_km,grid_reso)
            # Area limits
            latlims, lonlims = (np.min(latgrid), np.max(latgrid)), (np.min(longrid), np.max(longrid))
            # Read wind and focut to the area to speed up the interpolation
            wind = read_gnfra_wind(wind_folder,date,latlims=latlims,lonlims=lonlims,dataset=wind_product)
            # Gridded eastward and northward fluxes
            flux_xr = gridded_flux(data_path,wind,date,latgrid,longrid,latgrid_km,longrid_km,
                                   sats=sats,wind_product=wind_product,
                                   co2_noise_scenario=co2_noise_scenario,
                                   co2_cloud_threshold=co2_cloud_threshold,
                                   no2_noise_scenario=no2_noise_scenario,
                                   no2_cloud_threshold=no2_cloud_threshold)
            # Concatenate the data from the remaining dates to the flux_xr along 'n_images' dimension
            if len(dates) > 1:
                for date in dates[1:]:
                    date = date.date()
                    wind = read_gnfra_wind(wind_folder,date,latlims=latlims,lonlims=lonlims,dataset=wind_product)
                    flux_xr_next = gridded_flux(data_path,wind,date,latgrid,longrid,latgrid_km,longrid_km,
                                                sats=sats,wind_product=wind_product,
                                                co2_noise_scenario=co2_noise_scenario,
                                                co2_cloud_threshold=co2_cloud_threshold,
                                                no2_noise_scenario=no2_noise_scenario,
                                                no2_cloud_threshold=no2_cloud_threshold,
                                                specie=gas)
                    flux_xr = xr.concat((flux_xr,flux_xr_next), 'n_images')
            # Compute divergence map from the temporally averaged fluxes
            div_xr = compute_divergence_map(flux_xr,specie=gas)
            end_time = time.process_time()
            print(f'...done in {end_time-start_time} seconds')
            print(f"Optimizing trace gas {gas} emissions for source '{source_name}'...")
            start_time = time.process_time()
            fitting_xr = peak_fitting(div_xr,specie=gas,AM_options=AM_options)
            end_time = time.process_time()
            print(f'...done in {end_time-start_time} seconds')
            # Emission estimation and standard deviation in g/s
            emis_est, emis_std = fitting_xr['opt'].data[0], np.sqrt(fitting_xr['cov'].data[0,0])
            # Conversion to Mt/y
            emis_est, emis_std = emis_est*3600*24*365.24219/10**12, emis_std*3600*24*365.24219/10**12
            # Write results to xarray dataset
            results_ds[f'{gas}_estimated_emissions'].loc[dict(source=source_name)] = emis_est
            results_ds[f'{gas}_estimated_emissions_std'].loc[dict(source=source_name)] = emis_std
    estimation_end = time.process_time()      
    print(f'Estimations done in {estimation_end-estimation_start} seconds')
    return results_ds

def peak_fitting(div_xr,specie='CO2',pixels=4,AM_options=None):
    latgrid_km, longrid_km = div_xr['latgrid_km'].data, div_xr['longrid_km'].data
    # Create pixels x pixels sized kilometer grid around the source
    row_slice = slice((len(latgrid_km)-1)//2-pixels,(len(latgrid_km)+1)//2+pixels,1)
    col_slice = slice((len(longrid_km)-1)//2-pixels,(len(longrid_km)+1)//2+pixels,1)
    # Corresponding 2d X and Y kilometer grids around the source
    Y, X = np.meshgrid(latgrid_km[row_slice],longrid_km[col_slice],indexing='ij')
    # Intial values
    if specie == 'NO2':
        init = [900, 5, 6, 0.0, 0.0, 0.0, -0.5]
        bounds = np.array([(0,2000), (1,9), (1,9), (-10,10), (-10,10), (-1,1), (-1,1)])
        scales = [1e3,1e0,1e0,1e0,1e0,1e0,1e-1]
    elif specie == 'CO2':
        init = [9e5, 7, 7, 0.0, 0.0, 0.0, -500]
        bounds = np.array([(0,2e6), (1,9), (1,9), (-10,10), (-10,10), (-1,1), (-1000,1000)])
        scales = [1e5,1e0,1e0,1e0,1e0,1e0,1e2]
    # Divergence near the source
    div_map = div_xr[f'{specie}_div'].data[row_slice, col_slice]
    # Residual vector or component-wise difference between the model and the data
    residual = lambda params: model_residual(params,(X,Y),div_map)
    # Least square optimization using scipy with default Trust Region Reflective algorithm
    optimized = least_squares(residual, init, x_scale=scales, bounds=bounds.T, method='trf')
    opt = optimized.x
    J = optimized.jac
    cov = np.linalg.inv(J.T @ J)
    # Optimization using adaptive metropolis algorithm if AM_options are given
    if AM_options is not None:
        warmup, samples = AM_options['warmup'], AM_options['samples']
        sampling = AM_MC(residual,opt,samples,warmup,C0=cov,
                         bounds=bounds,init_fit=False,progress=False)
        opt, cov = sampling['MAP'].data, sampling['cov'].data
    names = ['peak_integral','sigma_x','sigma_y','x0','y0','corr','background']
    ds = xr.Dataset({'opt': xr.DataArray(opt,dims=['params'],coords=dict(params=names)),
                     'cov': xr.DataArray(cov,dims=['params','params'],coords=dict(params=names)),
                     'bg': opt[-1]})
    return ds